#define _XOPEN_SOURCE 1
#define _XOPEN_SOURCE_EXTENDED 1
#include <unistd.h>
#include <limits.h>
#include <stddef.h>

static char const prog[]="caldelay";

static char const var_weekday[]="$WEEKDAY";
static char const var_year[]="$YEAR";
static char const var_month[]="$MONTH";
static char const var_monthday[]="$MONTHDAY";
static char const var_hour[]="$HOUR";
static char const var_minute[]="$MINUTE";
static char const var_second[]="$SECOND";

static int increased=0;
static int need_increase=0;
struct caltime ct;

static unsigned long limit_weekday(void) { return  7; }
static unsigned long limit_year   (void) { return -1; }
static unsigned long limit_month  (void) { return 12; }
static unsigned long limit_monthday(void) {
  static unsigned long const limits[]={
    31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
  };
  unsigned long leap;
  if (ct.date.month!=1)
    return limits[ct.date.month];
  leap=ct.date.year;
  if (leap%4!=0) leap=0;
  else {
    leap/=4;
    if (leap%25!=0) leap=1;
    else {
      leap/=25;
      if (leap%4!=0) leap=0;
      else leap=1;
    }
  }
  return 28+leap;
}
static unsigned long limit_hour  (void) { return 24; }
static unsigned long limit_minute(void) { return 60; }
static unsigned long limit_second(void) { return 60; }

typedef enum { eq, div } spectype;
typedef struct {
  char const* const env;
  spectype type;
  unsigned long spec;
  unsigned long value;
  unsigned long (* const limit)(void);
} unit;
static unit weekday_units[]={
  { var_weekday,  -1, -1, -1, limit_weekday  },
  { var_hour,     -1, -1, -1, limit_hour     },
  { var_minute,   -1, -1, -1, limit_minute   },
  { var_second,   -1, -1, -1, limit_second   }
};
static unit monthday_units[]={
  { var_year,     -1, -1, -1, limit_year     },
  { var_month,    -1, -1, -1, limit_month    },
  { var_monthday, -1, -1, -1, limit_monthday },
  { var_hour,     -1, -1, -1, limit_hour     },
  { var_minute,   -1, -1, -1, limit_minute   },
  { var_second,   -1, -1, -1, limit_second   }
};

static void do_unit(unit* const u) {
  unsigned long value=(increased? 0: u->value);
  unsigned long const spec=u->spec;
  spectype const type=u->type;
  if (type==eq) {
    if (need_increase) return;
    if (value==spec) goto setval;
    if (value>spec) goto need_increase;
    value=spec;
  } else /* type==div */ {
    if (need_increase) {
      value+=spec;
      if (value>=u->limit()) return;
      increased=1;
      need_increase=0;
      goto setval;
    }
    if (value==0) goto setval;
    if (spec==0) goto need_increase;
    if (value%spec==0) goto setval;
    value+=spec;
    value-=value%spec;
    if (value>=u->limit()) goto need_increase;
  }
  increased=1;
  setval:
  u->value=value;
  return;
  need_increase:
  need_increase=1;
  return;
}

int main(int argc, char** argv) {
  struct tai now;
  int weekday;
  char const* x;
  unit* units;
  size_t n_units;
  size_t i;

  if (argc<2)
    die4x(111, prog, usage, prog, " program [arg ...]");

  x=env_get(var_weekday+1);
  if (x!=null) {
    units=weekday_units;
    n_units=sizeof weekday_units/sizeof (unit);
  } else {
    units=monthday_units;
    n_units=sizeof monthday_units/sizeof (unit);
  }

  {
    int seen_any=0;
    for (i=0; i!=n_units; ++i) {
      x=env_get(units[i].env+1);
      if (x!=null) {
        unsigned int len;
        seen_any=1;
        switch (*x) {
          case '=': units[i].type=eq;  break;
          case '/': units[i].type=div; break;
          default:
            die4x(111, prog, fatal, units[i].env, err_malformed);
        }
        ++x;
        len=scan_ulong(x, &units[i].spec);
        if (len==0 || x[len]!='\0')
          die4x(111, prog, fatal, units[i].env, err_malformed);
      } else if (seen_any) {
        units[i].type=eq;
        units[i].spec=0;
      } else {
        units[i].type=div;
        units[i].spec=1;
      }
    }
  }

  x=env_get(var_delay+1);
  if (x!=null) {
    {
      struct tai d;
      {
        unsigned char taibuf[TAI_PACK]={ 0, 0, 0, 0, 0, 0, 0, 0 };
        {
          unsigned char* bufp;
          unsigned long delay;
          {
            unsigned int const len=scan_ulong(x, &delay);
            if (len==0 || x[len]!='\0')
              die4x(111, prog, fatal, var_delay, err_malformed);
          }
          bufp=taibuf+sizeof taibuf;
          while (delay!=0) {
            if (bufp==taibuf)
              die5x(111, prog, fatal, var_delay, err_oflow, "struct tai");
            --bufp;
            *bufp=delay&(unsigned long)0xff;
            delay>>=8;
          }
        }
        tai_unpack(taibuf, &d);
      }
      tai_now(&now);
      tai_add(&now, &now, &d);
      caltime_utc(&ct, &now, &weekday, null);
      tai_sub(&now, &now, &d);
    }

    if (units==weekday_units) {
      units[0].value=weekday;
      units[1].value=ct.hour;
      units[2].value=ct.minute;
      units[3].value=ct.second;
    } else {
      units[0].value=ct.date.year;
      units[1].value=ct.date.month-1;
      units[2].value=ct.date.day-1;
      units[3].value=ct.hour;
      units[4].value=ct.minute;
      units[5].value=ct.second;
    }

    for (i=0; i!=n_units;) {
      do_unit(&units[i]);
      while (need_increase) {
        if (i==0) {
          if (units==monthday_units)
            die3x(111, prog, fatal, "all run times are in the past");
          units[0].value=0;
          ct.date.day+=7;
          continue;
        }
        --i;
        do_unit(&units[i]);
      }
      ++i;
    }

    if (units==weekday_units) {
      ct.date.day +=units[0].value;
      ct.date.day -=weekday;
      ct.hour      =units[1].value;
      ct.minute    =units[2].value;
      ct.second    =units[3].value;
    } else {
      ct.date.year =units[0].value;
      ct.date.month=units[1].value+1;
      ct.date.day  =units[2].value+1;
      ct.hour      =units[3].value;
      ct.minute    =units[4].value;
      ct.second    =units[5].value;
    }

    {
      char delaybuf[(sizeof (unsigned long)*CHAR_BIT+2)/3+1];
      {
        unsigned long delay;
        {
          unsigned char taibuf[TAI_PACK];
          {
            struct tai then;
            caltime_tai(&ct, &then);
            tai_sub(&then, &then, &now);
            tai_pack(taibuf, &then);
          }
          delay=0;
          for (i=0; i<TAI_PACK; ++i) {
            unsigned long const d=delay<<8;
            if (d>>8!=delay)
              die5x(111, prog, fatal, var_delay, err_oflow, t_long);
            delay=d+taibuf[i];
            if (delay<d)
              die5x(111, prog, fatal, var_delay, err_oflow, t_long);
          }
        }
        delaybuf[fmt_ulong(delaybuf, delay)]='\0';
      }
      if (pathexec_env(var_delay+1, delaybuf)==0)
        die6sys(111, prog, fatal, err_unable, err_setvar, var_delay, err_);
    }
  }

  ++argv;
  pathexec(argv);
  die6sys(111, prog, fatal, err_unable, err_exec, argv[0], err_);
}
